'use strict';

var p = require('path');
// const printAST = require('ast-pretty-print')

var macrosRegex = /[./]macro(\.js)?$/;

module.exports = macrosPlugin;

function macrosPlugin(babel) {
  return {
    name: 'macros',
    visitor: {
      ImportDeclaration(path, state) {
        var isMacros = looksLike(path, {
          node: {
            source: {
              value: function value(v) {
                return macrosRegex.test(v);
              }
            }
          }
        });
        if (!isMacros) {
          return;
        }
        var imports = path.node.specifiers.map(function (s) {
          return {
            localName: s.local.name,
            importedName: s.type === 'ImportDefaultSpecifier' ? 'default' : s.imported.name
          };
        });
        var source = path.node.source.value;
        applyMacros({ path, imports, source, state, babel });
        path.remove();
      },
      CallExpression(path, state) {
        var isMacros = looksLike(path, {
          node: {
            callee: {
              type: 'Identifier',
              name: 'require'
            },
            arguments: function _arguments(args) {
              return args.length === 1 && macrosRegex.test(args[0].value);
            }
          },
          parent: {
            type: 'VariableDeclarator'
          }
        });
        if (!isMacros) {
          return;
        }
        var name = path.parent.id.name;
        var source = path.node.arguments[0].value;
        applyMacros({
          path,
          imports: [{ localName: name, importedName: 'default' }],
          source,
          state,
          babel
        });
        path.parentPath.remove();
      }
    }
  };
}

function applyMacros(_ref) {
  var path = _ref.path,
      imports = _ref.imports,
      source = _ref.source,
      state = _ref.state,
      babel = _ref.babel;
  var filename = state.file.opts.filename;

  var hasReferences = false;
  var referencePathsByImportName = imports.reduce(function (byName, _ref2) {
    var importedName = _ref2.importedName,
        localName = _ref2.localName;

    byName[importedName] = path.scope.getBinding(localName).referencePaths;
    hasReferences = hasReferences || Boolean(byName[importedName].length);
    return byName;
  }, {});
  if (!hasReferences) {
    return;
  }
  var requirePath = source;
  var isRelative = source.indexOf('.') === 0;
  if (isRelative) {
    requirePath = p.join(p.dirname(getFullFilename(filename)), source);
  }
  // eslint-disable-next-line import/no-dynamic-require
  var macros = require(requirePath);
  macros({
    references: referencePathsByImportName,
    state,
    babel
  });
}

/*
 istanbul ignore next
 because this is hard to test
 and not worth it...
 */
function getFullFilename(filename) {
  if (p.isAbsolute(filename)) {
    return filename;
  }
  return p.join(process.cwd(), filename);
}

function looksLike(a, b) {
  return a && b && Object.keys(b).every(function (bKey) {
    var bVal = b[bKey];
    var aVal = a[bKey];
    if (typeof bVal === 'function') {
      return bVal(aVal);
    }
    return isPrimitive(bVal) ? bVal === aVal : looksLike(aVal, bVal);
  });
}

function isPrimitive(val) {
  // eslint-disable-next-line
  return val == null || /^[sbn]/.test(typeof val);
}